use parking_lot;

use youtube::convert_file_name;

use serenity;
use serenity::client::CACHE;
use serenity::model::id::{ChannelId, GuildId};
use serenity::voice::{AudioSource, Handler, LockedAudio};

use std::cell::RefCell;
use std::ops::DerefMut;
use std::sync::{Arc, Mutex, MutexGuard};
use std::thread;
use std::thread::JoinHandle;
use std::time;

macro_rules! check_option {
    ($v:expr) => {
        match $v {
            Some(t) => t,
            None => return,
        }
    };
}

pub struct Song {
    pub source: Box<AudioSource>,
    pub name: String,
}

pub struct MediaData {
    playlist: RefCell<Vec<Song>>,
    current_song: Mutex<RefCell<Option<LockedAudio>>>,
    watcher_thread: RefCell<Option<JoinHandle<()>>>,
}

fn guild_id(channel_id: ChannelId) -> Option<GuildId> {
    match CACHE.read().guild_channel(channel_id) {
        Some(channel) => Some(channel.read().guild_id),
        None => {
            super::check_msg(channel_id.say("Error finding channel info"));
            None
        }
    }
}

fn handler<'a>(
    guild_id: GuildId,
    manager: &'a mut parking_lot::MutexGuard<
        '_,
        serenity::client::bridge::voice::ClientVoiceManager,
    >,
) -> Option<&'a mut Handler> {
    manager.get_mut(guild_id)
}

impl MediaData {
    fn song_mut(&self) -> MutexGuard<RefCell<Option<LockedAudio>>> {
        self.current_song.lock().unwrap()
    }

    fn reset_thread(&self) {
        *self.watcher_thread.borrow_mut() = None;
    }

    fn start_thread(
        media: &Arc<MediaData>,
        channel_id: ChannelId,
        manager_lock: &Arc<
            serenity::prelude::Mutex<serenity::client::bridge::voice::ClientVoiceManager>,
        >,
    ) {
        let media_clone = media.clone();
        let manager_clone = manager_lock.clone();

        let mut watcher_mut = media.watcher_thread.borrow_mut();

        if watcher_mut.is_none() {
            *watcher_mut = Some(thread::spawn(move || loop {
                {
                    let song_lock = media_clone.song_mut();
                    let mut borrow = song_lock.borrow_mut();

                    if let Some(ref mut song) = borrow.deref_mut() {
                        let song_lock = song.clone();
                        let audio = song_lock.lock();

                        if audio.finished {
                            if !Self::next_song(
                                song,
                                media_clone.playlist.borrow_mut().deref_mut(),
                                channel_id,
                                &manager_clone,
                            ) {
                                manager_clone
                                    .lock()
                                    .remove(check_option!(guild_id(channel_id)));

                                media_clone.reset_thread();

                                println!("left channel");

                                return;
                            }
                        }

                        continue;
                    }

                    let mut playlist = media_clone.playlist.borrow_mut();

                    if !playlist.is_empty() {
                        let mut manager = manager_clone.lock();

                        if let Some(handler) =
                            handler(check_option!(guild_id(channel_id)), &mut manager)
                        {
                            let first = playlist.remove(0);

                            *borrow = Some(handler.play_returning(first.source));

                            super::check_msg(channel_id.say(format!(
                                "Playing song: {}",
                                super::youtube::convert_file_name(first.name)
                            )));
                        } else {
                            return;
                        }
                    } else {
                        manager_clone
                            .lock()
                            .remove(check_option!(guild_id(channel_id)));

                        media_clone.reset_thread();

                        println!("left channel");

                        return;
                    }
                }

                let five_sec = time::Duration::from_secs(5);
                thread::sleep(five_sec);
            }));
        }
    }

    fn next_song(
        song: &mut LockedAudio,
        playlist: &mut Vec<Song>,
        channel_id: ChannelId,
        manager_lock: &Arc<
            serenity::prelude::Mutex<serenity::client::bridge::voice::ClientVoiceManager>,
        >,
    ) -> bool {
        let mut manager = manager_lock.lock();

        let guild_id = match guild_id(channel_id) {
            Some(id) => id,
            None => return false,
        };

        if let Some(handler) = handler(guild_id, &mut manager) {
            if playlist.is_empty() {
                return false;
            } else {
                let first = playlist.remove(0);

                *song = handler.play_returning(first.source);

                super::check_msg(channel_id.say(format!(
                    "Playing song: {}",
                    super::youtube::convert_file_name(first.name)
                )));

                return true;
            }
        }

        return false;
    }
}

impl Default for MediaData {
    fn default() -> MediaData {
        MediaData {
            playlist: RefCell::new(Vec::new()),
            current_song: Mutex::new(RefCell::new(None)),
            watcher_thread: RefCell::new(None),
        }
    }
}

unsafe impl Send for MediaData {}
unsafe impl Sync for MediaData {}

pub struct Play {
    media: Arc<MediaData>,
}

impl Play {
    pub fn new(media_data: Arc<MediaData>) -> Play {
        Play { media: media_data }
    }

    fn check_for_continue(song_lock: MutexGuard<RefCell<Option<LockedAudio>>>) -> bool {
        match song_lock.borrow_mut().deref_mut() {
            Some(song) => {
                let song_clone = song.clone();
                let mut audio_lock = song_clone.lock();

                audio_lock.play();
                true
            }
            None => false,
        }
    }

    fn check_join_channel(
        manager_lock: &Arc<parking_lot::Mutex<serenity::client::bridge::voice::ClientVoiceManager>>,
        msg: &serenity::model::channel::Message,
    ) {
        let guild = match msg.guild() {
            Some(guild) => guild,
            None => {
                display_error!(msg.channel_id.say("Groups and DMs not supported"));

                return;
            }
        };

        let guild_id = guild.read().id;

        let channel_id = guild
            .read()
            .voice_states
            .get(&msg.author.id)
            .and_then(|voice_state| voice_state.channel_id);

        let connect_to = match channel_id {
            Some(channel) => channel,
            None => {
                display_error!(msg.reply("Not in a voice channel"));

                return;
            }
        };

        let mut manager = manager_lock.lock();

        manager.join(guild_id, connect_to);
    }
}

impl serenity::framework::standard::Command for Play {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        mut ctx: &mut serenity::client::Context,
        msg: &serenity::model::channel::Message,
        mut args: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        let url = match args.single::<String>() {
            Ok(url) => {
                if !url.starts_with("http") {
                    super::check_msg(msg.channel_id.say("Must provide a valid URL"));

                    return Ok(());
                }

                url
            }
            Err(_) => {
                if !Self::check_for_continue(self.media.song_mut()) {
                    super::check_msg(msg.channel_id.say("Must provide a URL to a video or audio"));
                }

                return Ok(());
            }
        };

        let mut source = match super::youtube::youtube_dl(&url) {
            Ok(source) => source,
            Err(why) => {
                println!("Err starting source: {:?}", why);

                super::check_msg(msg.channel_id.say("Error sourcing ffmpeg"));

                return Ok(());
            }
        };

        self.media.playlist.borrow_mut().append(&mut source);

        let mut manager_lock = ctx.data
            .lock()
            .get::<super::VoiceManager>()
            .cloned()
            .unwrap();

        Self::check_join_channel(&manager_lock, msg);

        MediaData::start_thread(&self.media, msg.channel_id, &manager_lock);

        Ok(())
    }
}

pub struct Pause {
    media: Arc<MediaData>,
}

impl Pause {
    pub fn new(media_data: Arc<MediaData>) -> Pause {
        Pause { media: media_data }
    }
}

impl serenity::framework::standard::Command for Pause {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        _: &mut serenity::client::Context,
        _: &serenity::model::channel::Message,
        _: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        let song_lock = self.media.song_mut();

        if let Some(song) = song_lock.borrow_mut().deref_mut() {
            let song_clone = song.clone();
            let mut audio_lock = song_clone.lock();

            audio_lock.pause();
        }

        Ok(())
    }
}

pub struct List {
    media: Arc<MediaData>,
}

impl List {
    pub fn new(media_data: Arc<MediaData>) -> List {
        List { media: media_data }
    }
}

impl serenity::framework::standard::Command for List {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        _: &mut serenity::client::Context,
        msg: &serenity::model::channel::Message,
        _: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        let mut output = String::new();

        let playlist = self.media.playlist.borrow();

        output += &format!(
            "{} {} queued\n",
            playlist.len(),
            if playlist.len() == 1 { "song" } else { "songs" }
        );

        for (i, song) in playlist.iter().enumerate() {
            output += &format!("\t{}.\t{}\n", i + 1, convert_file_name(song.name.clone()));
        }

        super::check_msg(msg.channel_id.say(output));

        Ok(())
    }
}

pub struct Help {
    prefix: String,
    commands: Vec<String>,
}

impl Help {
    pub fn new(prefix: &String, commands: Vec<String>) -> Help {
        Help {
            prefix: prefix.clone(),
            commands: commands,
        }
    }
}

impl serenity::framework::standard::Command for Help {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        _: &mut serenity::client::Context,
        msg: &serenity::model::channel::Message,
        _: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        let mut output = String::new();

        output += "Available commands:\n";

        for command in &self.commands {
            output += &format!("\t{}{}\n", self.prefix, command);
        }

        super::check_msg(msg.channel_id.say(output));

        Ok(())
    }
}

pub struct Stop {
    media: Arc<MediaData>,
}

impl Stop {
    pub fn new(media_data: Arc<MediaData>) -> Stop {
        Stop { media: media_data }
    }
}

impl serenity::framework::standard::Command for Stop {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        ctx: &mut serenity::client::Context,
        msg: &serenity::model::channel::Message,
        _: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        self.media.playlist.borrow_mut().clear();

        {
            let song = self.media.song_mut();
            *song.borrow_mut() = None;
        }

        {
            let mut manager_lock = ctx.data
                .lock()
                .get::<super::VoiceManager>()
                .cloned()
                .unwrap();

            let mut manager = manager_lock.lock();
            let guild_id = match guild_id(msg.channel_id) {
                Some(guild_id) => guild_id,
                None => return Ok(()),
            };

            let handler = match handler(guild_id, &mut manager) {
                Some(handler) => handler,
                None => return Ok(()),
            };

            println!("stopped handler");

            handler.stop();
        }

        Ok(())
    }
}
